@@ -18,8 +18,8 @@ use std::cmp::Ordering;
 use std::ffi::{c_void, CStr};
 use std::os::raw::c_char;
 use std::rc::Rc;
-use std::slice;
-use std::time::Duration;
+use std::time::{Duration, Instant};
+use std::{slice, thread};
 use windows::Win32::Foundation::{
     CloseHandle, GetLastError, HANDLE, HWND, LPARAM, WIN32_ERROR, WPARAM,
 };
@@ -44,6 +44,7 @@ pub const IRSDK_UNLIMITED_TIME: f64 = 604800.0;
 
 /// Client is main entry point into the library. Create a client and then you can
 /// access Sessions that have the telemetry data in then.
+#[derive(Debug)]
 pub struct Client {
     conn: Option<Rc<Connection>>,
     // Incremented each time we issue a new session. Allows for session to determine its expired even if
@@ -51,7 +52,7 @@ pub struct Client {
     session_id: i32,
 }
 impl Client {
-    /// creates a new Client
+    /// Creates a new Client
     pub fn new() -> Client {
         Client {
             conn: None,
@@ -75,7 +76,7 @@ impl Client {
     }
     /// When iRacing is running, then a Session will be available that contains the
     /// telemetry data. When iRacing is not running, then this returns None.
-    /// see also wait_for_session (When it exists)
+    /// See also wait_for_session.
     ///
     /// # Safety
     /// Creating a session requires dealing with memory mapped files and c strucutres
@@ -90,18 +91,50 @@ impl Client {
             let mut s = Session {
                 session_id: sid,
                 conn: self.conn.as_ref().unwrap().clone(),
-                last_tick_count: 0,
+                last_tick_count: -2,
                 data: bytes::BytesMut::new(),
                 expired: false,
             };
-            if DataUpdateResult::Updated == s.get_new_data() {
+            let d = s.wait_for_data(Duration::from_millis(16));
+            if DataUpdateResult::Updated == d {
                 Some(s)
             } else {
                 None
             }
         }
     }
-    // TODO wait_for_session()
+    /// Will wait upto the the supplied wait duration for a iRacing session to be available.
+    /// If the wait timeout's returns None, otherwise returns the new Session.
+    /// Wait must fit into a 32bit number of milliseconds, about 49 days. otherwise it'll panic.
+    ///
+    /// # Safety
+    /// Creating a session requires dealing with memory mapped files and c strucutres
+    /// that are in them. A mismatch between our definition of the struct and iRacings
+    /// definition could cause chaos.
+    pub unsafe fn wait_for_session(&mut self, wait: Duration) -> Option<Session> {
+        // If we had a prior connection then we can use the new data event to wait for the
+        // session. Otherwise we need to poll session fn above.
+        match &self.conn {
+            Some(c) => {
+                c.wait_for_new_data(wait);
+                self.session()
+            }
+            None => {
+                let start = Instant::now();
+                let loop_wait = Duration::from_millis(1000);
+                loop {
+                    let r = self.session();
+                    if r.is_some() || start.elapsed() > wait {
+                        return r;
+                    }
+                    // It takes iRacing a few seconds to get from when you can connect
+                    // to the telemetry to when you can actually do anything in the simulator
+                    // so a once a second check seems like plenty.
+                    thread::sleep(loop_wait);
+                }
+            }
+        }
+    }
 }
 impl Default for Client {
     fn default() -> Self {
@@ -253,7 +286,14 @@ impl Session {
             var.session_id, self.session_id,
             "programmer error, Var was issued by a different Session"
         );
-        let x = self.data.as_ptr().add(var.hdr.offset as usize);
+        // verify that value is inside the buffer
+        let var_offset = var.hdr.offset as usize;
+        assert!(
+            var.size() + var_offset <= self.data.len(),
+            "The value appears to be outside the buffer"
+        );
+
+        let x = self.data.as_ptr().add(var_offset);
         if var.hdr.count == 1 {
             match var.hdr.var_type {
                 VarType::Char => Value::Char(*x),
@@ -368,6 +408,10 @@ impl Var {
     pub fn count(&self) -> usize {
         self.hdr.count as usize
     }
+    /// returns the size of a value for this Var in bytes.
+    fn size(&self) -> usize {
+        self.var_type().size() * self.count()
+    }
 }
 impl fmt::Debug for Var {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
@@ -488,7 +532,9 @@ impl Drop for Connection {
         }
     }
 }
+
 #[repr(C)]
+#[derive(Debug)]
 struct IrsdkBuf {
     tick_count: i32, // used to detect changes in data
     buf_offset: i32, // offset from header
@@ -496,6 +542,7 @@ struct IrsdkBuf {
 }
 
 #[repr(C)]
+#[derive(Debug)]
 struct IrsdkHeader {
     ver: i32,                   // this api header version, see IRSDK_VER
     status: flags::StatusField, // bitfield using irsdk_StatusField
@@ -586,7 +633,19 @@ pub enum VarType {
     #[deprecated]
     Etcount = 6,
 }
-
+impl VarType {
+    fn size(&self) -> usize {
+        match *self {
+            VarType::Char => 1,
+            VarType::Bool => 1,
+            VarType::Int => 4,
+            VarType::Bitfield => 4,
+            VarType::Float => 4,
+            VarType::Double => 8,
+            _ => todo!(), //Etcount
+        }
+    }
+}
 /// An instance of a value for a variable.
 #[derive(Debug, Clone, Copy, PartialEq)]
 pub enum Value<'a> {
@@ -786,6 +845,10 @@ impl FromValue for flags::PaceFlags {
 #[cfg(test)]
 mod tests {
 
+    use std::ptr;
+
+    use crate::flags::StatusField;
+
     use super::*;
 
     #[test]
@@ -809,4 +872,213 @@ mod tests {
         assert_eq!(m(-1, 0), 0x0000FFFF);
         assert_eq!(m(0x1234, 0x5678), 0x56781234);
     }
+
+    #[test]
+    fn test_var_size() {
+        let f = |t, c| {
+            Var {
+                session_id: 0,
+                hdr: IrsdkVarHeader {
+                    var_type: t,
+                    offset: 0,
+                    count: c,
+                    count_as_time: 0,
+                    pad: [0, 0, 0], // (16 byte align)
+                    name: [0; IRSDK_MAX_STRING],
+                    desc: [0; IRSDK_MAX_DESC],
+                    unit: [0; IRSDK_MAX_STRING],
+                },
+            }
+        };
+        assert_eq!(512, f(VarType::Double, 64).size());
+        assert_eq!(24, f(VarType::Double, 3).size());
+        assert_eq!(8, f(VarType::Double, 1).size());
+        assert_eq!(12, f(VarType::Float, 3).size());
+        assert_eq!(4, f(VarType::Float, 1).size());
+        assert_eq!(12, f(VarType::Int, 3).size());
+        assert_eq!(4, f(VarType::Int, 1).size());
+        assert_eq!(12, f(VarType::Bitfield, 3).size());
+        assert_eq!(4, f(VarType::Bitfield, 1).size());
+        assert_eq!(3, f(VarType::Char, 3).size());
+        assert_eq!(1, f(VarType::Char, 1).size());
+        assert_eq!(3, f(VarType::Bool, 3).size());
+        assert_eq!(1, f(VarType::Bool, 1).size());
+    }
+
+    #[test]
+    fn test_irsdk_var_header() {
+        let mut h = IrsdkVarHeader {
+            var_type: VarType::Float,
+            offset: 32,
+            count: 1,
+            count_as_time: 0,
+            pad: [0; 3],
+            name: [0; IRSDK_MAX_STRING],
+            desc: [0; IRSDK_MAX_DESC],
+            unit: [0; IRSDK_MAX_STRING],
+        };
+        // there must be an easier way than this
+        h.name[0] = 'b' as u8;
+        h.name[1] = 'o' as u8;
+        h.name[2] = 'b' as u8;
+        assert_eq!(Ok("bob"), h.name());
+        assert!(h.has_name("bob"));
+        assert!(!h.has_name("alice"));
+        assert!(!h.has_name("bobby"));
+    }
+
+    #[test]
+    fn test_var_value() {
+        let b = || IrsdkBuf {
+            tick_count: 1,
+            buf_offset: 0,
+            pad: [0, 2],
+        };
+        let mut h = IrsdkHeader {
+            ver: 2,
+            status: StatusField::CONNECTED,
+            tick_rate: 60,
+            session_info_update: 1,
+            session_info_len: 0,
+            session_info_offset: 100,
+            num_vars: 0,
+            var_header_offset: 0,
+            num_buf: 3,
+            buf_len: 12,
+            pad1: [0; 2],
+            var_buf: [b(), b(), b(), b()],
+        };
+        let mut s = Session {
+            session_id: 1,
+            conn: Rc::new(Connection {
+                file_mapping: HANDLE::default(),
+                shared_mem: ptr::null_mut(),
+                header: ptr::addr_of_mut!(h),
+                new_data: HANDLE::default(),
+                broadcast_msg_id: 1,
+            }),
+            last_tick_count: 1,
+            data: bytes::BytesMut::new(),
+            expired: false,
+        };
+        // char/bool
+        s.data.extend_from_slice(&[55, 56, 57, 58, 1, 0, 1, 0]);
+        // int
+        s.data.extend_from_slice(&[1, 2, 3, 4, 5, 6, 7, 8]);
+        // float
+        s.data
+            .extend_from_slice(&[0x00, 0x00, 0x80, 0x3f, 0, 0, 0, 0xc0]);
+        // double
+        s.data
+            .extend_from_slice(&[0, 0, 0, 0, 0, 0, 0, 0xc0, 2, 0, 0, 0, 0, 0, 0xF0, 0x3F]);
+        let v = |t, o, c| Var {
+            hdr: IrsdkVarHeader {
+                var_type: t,
+                offset: o,
+                count: c,
+                count_as_time: 0,
+                pad: [0; 3],
+                name: [0; IRSDK_MAX_STRING],
+                desc: [0; IRSDK_MAX_DESC],
+                unit: [0; IRSDK_MAX_STRING],
+            },
+            session_id: 1,
+        };
+        unsafe {
+            assert_eq!(s.var_value(&v(VarType::Char, 0, 1)), Value::Char(55));
+            assert_eq!(s.var_value(&v(VarType::Char, 1, 1)), Value::Char(56));
+            assert_eq!(s.var_value(&v(VarType::Bool, 4, 1)), Value::Bool(true));
+            assert_eq!(s.var_value(&v(VarType::Bool, 5, 1)), Value::Bool(false));
+            assert_eq!(s.var_value(&v(VarType::Int, 8, 1)), Value::Int(0x04030201));
+            assert_eq!(
+                s.var_value(&v(VarType::Bitfield, 8, 1)),
+                Value::Bitfield(0x04030201)
+            );
+            assert_eq!(s.var_value(&v(VarType::Float, 16, 1)), Value::Float(1.0));
+            assert_eq!(s.var_value(&v(VarType::Float, 20, 1)), Value::Float(-2.0));
+            assert_eq!(s.var_value(&v(VarType::Double, 24, 1)), Value::Double(-2.0));
+            assert_eq!(
+                s.var_value(&v(VarType::Double, 32, 1)),
+                Value::Double(1.0000000000000004)
+            );
+            assert_eq!(
+                s.var_value(&v(VarType::Char, 0, 3)),
+                Value::Chars(&[55, 56, 57])
+            );
+            assert_eq!(
+                s.var_value(&v(VarType::Bool, 4, 4)),
+                Value::Bools(&[true, false, true, false])
+            );
+            assert_eq!(
+                s.var_value(&v(VarType::Int, 8, 2)),
+                Value::Ints(&[0x04030201, 0x08070605])
+            );
+            assert_eq!(
+                s.var_value(&v(VarType::Bitfield, 8, 2)),
+                Value::Bitfields(&[0x04030201, 0x08070605])
+            );
+            assert_eq!(
+                s.var_value(&v(VarType::Float, 16, 2)),
+                Value::Floats(&[1.0, -2.0])
+            );
+            assert_eq!(
+                s.var_value(&v(VarType::Double, 24, 2)),
+                Value::Doubles(&[-2.0, 1.0000000000000004])
+            );
+        }
+    }
+    #[test]
+    #[should_panic]
+    fn test_cant_read_past_buffer() {
+        // ugh, need something better for setting these types of tests up.
+        let b = || IrsdkBuf {
+            tick_count: 1,
+            buf_offset: 0,
+            pad: [0, 2],
+        };
+        let mut h = IrsdkHeader {
+            ver: 2,
+            status: StatusField::CONNECTED,
+            tick_rate: 60,
+            session_info_update: 1,
+            session_info_len: 0,
+            session_info_offset: 100,
+            num_vars: 0,
+            var_header_offset: 0,
+            num_buf: 3,
+            buf_len: 12,
+            pad1: [0; 2],
+            var_buf: [b(), b(), b(), b()],
+        };
+        let mut s = Session {
+            session_id: 1,
+            conn: Rc::new(Connection {
+                file_mapping: HANDLE::default(),
+                shared_mem: ptr::null_mut(),
+                header: ptr::addr_of_mut!(h),
+                new_data: HANDLE::default(),
+                broadcast_msg_id: 1,
+            }),
+            last_tick_count: 1,
+            data: bytes::BytesMut::new(),
+            expired: false,
+        };
+        s.data.extend_from_slice(&[1, 2, 3, 4]);
+        let v = Var {
+            hdr: IrsdkVarHeader {
+                var_type: VarType::Int,
+                offset: 2,
+                count: 1,
+                count_as_time: 0,
+                pad: [0; 3],
+                name: [0; IRSDK_MAX_STRING],
+                desc: [0; IRSDK_MAX_DESC],
+                unit: [0; IRSDK_MAX_STRING],
+            },
+            session_id: 1,
+        };
+        unsafe {
+            s.var_value(&v);
+        }
+    }
 }
